@@ -1,14 +1,18 @@
 /*
-** Copyright 2000-2006 Double Precision, Inc.
+** Copyright 2000-2008 Double Precision, Inc.
 ** See COPYING for distribution information.
 */
 #include	"config.h"
 #include	"argparse.h"
 #include	"spipe.h"
+#define COURIERTCPD_EXPOSE_OPENSSL 1
 #include	"libcouriertls.h"
+#include	<openssl/rand.h>
+#include	<openssl/x509.h>
 #include	"tlscache.h"
 #include	"rfc1035/rfc1035.h"
 #include	"soxwrap/soxwrap.h"
+#include	"random128/random128.h"
 #ifdef  getc
 #undef  getc
 #endif
@@ -98,12 +102,14 @@ static int ssl_verify_callback(int goodcert, X509_STORE_CTX *x509)
 		X509_STORE_CTX_get_ex_data(x509,
 					   SSL_get_ex_data_X509_STORE_CTX_idx()
 					   );
-	const struct tls_info *info=SSL_get_app_data(ssl);
+	struct tls_info *info=SSL_get_app_data(ssl);
 
 	if (info->peer_verify_domain || get_peer_verify_level(info))
 	{
 		if (!goodcert)
 			return (0);
+
+		info->certificate_verified=1;
 	}
 
 	return (1);
@@ -363,6 +369,114 @@ static int process_certfile(SSL_CTX *ctx, const char *certfile, const char *ip,
 	return (*func)(ctx, certfile);
 }
 
+static int client_cert_cb(ssl_handle ssl, X509 **x509, EVP_PKEY **pkey)
+{
+	struct tls_info *info=(struct tls_info *)SSL_get_app_data(ssl);
+	int i;
+	const char *pem_cert;
+	size_t pem_cert_size;
+	STACK_OF(X509_NAME) *client_cas;
+	int cert_num=0;
+	int rc;
+
+	if (info->getpemclientcert4ca == NULL)
+		return 0;
+
+	rc=0;
+	client_cas=SSL_get_client_CA_list(ssl);
+
+	if (info->loadpemclientcert4ca)
+		(*info->loadpemclientcert4ca)(info->app_data);
+
+	for (cert_num=0; (*info->getpemclientcert4ca)(cert_num, &pem_cert,
+						      &pem_cert_size,
+						      info->app_data);
+	     ++cert_num)
+	{
+		BIO *certbio;
+		int err;
+		X509 *x;
+
+		ERR_clear_error();
+
+		certbio=BIO_new_mem_buf((void *)pem_cert, pem_cert_size);
+
+		if (!certbio)
+		{
+			rc= -1;
+			break;
+		}
+
+		x=PEM_read_bio_X509(certbio, x509, NULL, NULL);
+
+		if (!x)
+		{
+			BIO_free(certbio);
+			continue;
+		}
+
+		for (i=0; client_cas && i<client_cas->stack.num; i++)
+		{
+			X509_NAME *cert=(X509_NAME *)client_cas->stack.data[i];
+
+			if (X509_NAME_cmp(cert,
+					  x->cert_info->issuer) == 0)
+				break;
+		}
+
+		if (!client_cas || i >= client_cas->stack.num)
+		{
+			BIO_free(certbio);
+			continue;
+		}
+
+		while ((x=PEM_read_bio_X509(certbio, NULL,
+					    NULL, 0)) != NULL)
+		{
+			if (SSL_CTX_add_extra_chain_cert(SSL_get_SSL_CTX(ssl),
+							 x)
+			    != 1)
+			{
+				X509_free(x);
+				rc= -1;
+				break;
+			}
+		}
+
+		err = ERR_peek_last_error();
+		if (rc || ERR_GET_LIB(err) != ERR_LIB_PEM ||
+		    ERR_GET_REASON(err) != PEM_R_NO_START_LINE)
+		{
+			BIO_free(certbio);
+			continue;
+		}
+		BIO_free(certbio);
+
+		ERR_clear_error();
+
+		certbio=BIO_new_mem_buf((void *)pem_cert, pem_cert_size);
+
+		if (!certbio)
+		{
+			rc= -1;
+			break;
+		}
+
+		if (!PEM_read_bio_PrivateKey(certbio, pkey, NULL, NULL))
+		{
+			BIO_free(certbio);
+			continue;
+		}
+
+		BIO_free(certbio);
+		rc=1;
+		break;
+	}
+	ERR_clear_error();
+	(*info->releasepemclientcert4ca)(info->app_data);
+	return rc;
+}
+
 SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 {
 	SSL_CTX *ctx;
@@ -402,8 +516,25 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 		return (NULL);
 	}
 
-	SSL_load_error_strings();
-	SSLeay_add_ssl_algorithms();
+	{
+		static int first=1;
+
+		if (first)
+		{
+			first=0;
+			SSL_load_error_strings();
+			SSLeay_add_ssl_algorithms();
+
+			while (RAND_status() != 1)
+			{
+				const char *p=random128();
+				size_t l=strlen(p);
+
+				RAND_add(p, l, l/16);
+			}
+		}
+	}
+
 
 	info_copy=malloc(sizeof(struct tls_info));
 
@@ -415,10 +546,15 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 
 	memcpy(info_copy, info, sizeof(*info_copy));
 	info_copy->isserver=isserver;
+	info_copy->certificate_verified=0;
+
+	if (!protocol || !*protocol)
+		protocol="SSL23";
 
 	ctx=SSL_CTX_new(protocol && strcmp(protocol, "SSL2") == 0
 							? SSLv2_method():
-		protocol && strcmp(protocol, "SSL3") == 0 ? SSLv23_method():
+		protocol && strcmp(protocol, "SSL3") == 0 ? SSLv3_method():
+		protocol && strcmp(protocol, "SSL23") == 0 ? SSLv23_method():
 		TLSv1_method());
 
 	if (!ctx)
@@ -430,8 +566,10 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 	SSL_CTX_set_app_data(ctx, info_copy);
 	SSL_CTX_set_options(ctx, SSL_OP_ALL);
 
-	if (ssl_cipher_list)
-		SSL_CTX_set_cipher_list(ctx, ssl_cipher_list);
+	if (!ssl_cipher_list)
+		ssl_cipher_list="SSLv3:TLSv1:!SSLv2:HIGH:!LOW:!MEDIUM:!EXP:!NULL@STRENGTH";
+
+	SSL_CTX_set_cipher_list(ctx, ssl_cipher_list);
 	SSL_CTX_set_timeout(ctx, session_timeout);
 
 	info_copy->tlscache=NULL;
@@ -467,7 +605,8 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 			|| (!SSL_CTX_load_verify_locations(ctx, peer_cert_file,
 				peer_cert_dir)))
 		{
-			sslerror(info, peer_cert_dir, -1);
+			sslerror(info, peer_cert_file ?
+				 peer_cert_file:peer_cert_dir, -1);
 			tls_destroy(ctx);
 			return (0);
 		}
@@ -535,6 +674,8 @@ SSL_CTX *tls_create(int isserver, const struct tls_info *info)
 	}
 	SSL_CTX_set_verify(ctx, get_peer_verify_level(info),
 			   ssl_verify_callback);
+	if (!isserver)
+		SSL_CTX_set_client_cert_cb(ctx, client_cert_cb);
 	return (ctx);
 }
 
@@ -660,6 +801,8 @@ static SSL_SESSION *cache_get(SSL *ssl, unsigned char *id, int id_len,
 	fprintf(stderr, "INFO: TLSCACHE: session %s\n",
 		wi.ret ? "found":"not found");
 #endif
+	if (wi.ret)
+		SSL_set_session_id_context(ssl, id, id_len);
 	return wi.ret;
 }
 
@@ -869,6 +1012,14 @@ SSL *tls_connect(SSL_CTX *ctx, int fd)
 	return (ssl);
 }
 
+void tls_disconnect(SSL *ssl, int fd)
+{
+	fcntl(fd, F_SETFL, 0);
+	SSL_set_shutdown(ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
+	SSL_free(ssl);
+	ERR_remove_state(0);
+}
+
 /* --------------------------------------- */
 
 int	tls_transfer(struct tls_transfer_info *t, SSL *ssl, int fd,
@@ -1053,3 +1204,359 @@ int tls_connecting(SSL *ssl)
 	return info->accept_interrupted || info->connect_interrupted;
 }
 
+int tls_certificate_verified(ssl_handle ssl)
+{
+	struct tls_info *info=(struct tls_info *)SSL_get_app_data(ssl);
+
+	return info->certificate_verified;
+}
+
+#define MAXDOMAINSIZE	256
+
+static time_t asn1toTime(ASN1_TIME *asn1Time)
+{
+	struct tm tm;
+	int offset;
+
+	if (asn1Time == NULL || asn1Time->length < 13)
+		return 0;
+
+	memset(&tm, 0, sizeof(tm));
+
+#define N2(n)	((asn1Time->data[n]-'0')*10 + asn1Time->data[(n)+1]-'0')
+
+#define CPY(f,n) (tm.f=N2(n))
+
+	CPY(tm_year,0);
+
+	if(tm.tm_year < 50)
+		tm.tm_year += 100; /* Sux */
+
+	CPY(tm_mon, 2);
+	--tm.tm_mon;
+	CPY(tm_mday, 4);
+	CPY(tm_hour, 6);
+	CPY(tm_min, 8);
+	CPY(tm_sec, 10);
+
+	offset=0;
+
+	if (asn1Time->data[12] != 'Z')
+	{
+		if (asn1Time->length < 17)
+			return 0;
+
+		offset=N2(13)*3600+N2(15)*60;
+
+		if (asn1Time->data[12] == '-')
+			offset= -offset;
+	}
+
+#undef N2
+#undef CPY
+
+	return mktime(&tm)-offset;
+}
+
+
+static void dump_x509(X509 *x509,
+		      void (*dump_func)(const char *, int cnt, void *),
+		      void *dump_arg)
+{
+	X509_NAME *subj=X509_get_subject_name(x509);
+	int nentries, j;
+	time_t timestamp;
+	static const char gcc_shutup[]="%Y-%m-%d %H:%M:%S";
+
+	if (!subj)
+		return;
+
+	(*dump_func)("Subject:\n", -1, dump_arg);
+
+	nentries=X509_NAME_entry_count(subj);
+	for (j=0; j<nentries; j++)
+	{
+		const char *obj_name;
+		X509_NAME_ENTRY *e;
+		ASN1_OBJECT *o;
+		ASN1_STRING *d;
+
+		int dlen;
+		unsigned char *ddata;
+
+		e=X509_NAME_get_entry(subj, j);
+		if (!e)
+			continue;
+
+		o=X509_NAME_ENTRY_get_object(e);
+		d=X509_NAME_ENTRY_get_data(e);
+
+		if (!o || !d)
+			continue;
+
+		obj_name=OBJ_nid2sn(OBJ_obj2nid(o));
+
+		dlen=ASN1_STRING_length(d);
+		ddata=ASN1_STRING_data(d);
+	
+		(*dump_func)("   ", -1, dump_arg);
+		(*dump_func)(obj_name, -1, dump_arg);
+		(*dump_func)("=", 1, dump_arg);
+		(*dump_func)((const char *)ddata, dlen, dump_arg);
+		(*dump_func)("\n", 1, dump_arg);
+		
+	}
+	(*dump_func)("\n", 1, dump_arg);
+
+	timestamp=asn1toTime(X509_get_notBefore(x509));
+
+	if (timestamp)
+	{
+		struct tm *tm=localtime(&timestamp);
+		char buffer[500];
+
+		buffer[strftime(buffer, sizeof(buffer)-1, gcc_shutup,
+				tm)]=0;
+
+		(*dump_func)("Not-Before: ", -1, dump_arg);
+		(*dump_func)(buffer, -1, dump_arg);
+		(*dump_func)("\n", 1, dump_arg);
+	}
+
+	timestamp=asn1toTime(X509_get_notAfter(x509));
+	if (timestamp)
+	{
+		struct tm *tm=localtime(&timestamp);
+		char buffer[500];
+
+		buffer[strftime(buffer, sizeof(buffer)-1, gcc_shutup,
+				tm)]=0;
+
+		(*dump_func)("Not-After: ", -1, dump_arg);
+		(*dump_func)(buffer, -1, dump_arg);
+		(*dump_func)("\n", 1, dump_arg);
+	}
+}
+
+void tls_dump_connection_info(ssl_handle ssl,
+			      int server,
+			      void (*dump_func)(const char *, int cnt, void *),
+			      void *dump_arg)
+{
+	SSL_CIPHER *cipher;
+
+	{
+		STACK_OF(X509) *peer_cert_chain=SSL_get_peer_cert_chain(ssl);
+		int i;
+
+		if (server)
+		{
+			X509 *x=SSL_get_peer_certificate(ssl);
+
+			if (x)
+			{
+				dump_x509(x, dump_func, dump_arg);
+				X509_free(x);
+			}
+		}
+
+		for (i=0; peer_cert_chain && i<peer_cert_chain->stack.num; i++)
+			dump_x509((X509 *)peer_cert_chain->stack.data[i],
+				  dump_func, dump_arg);
+	}
+
+	cipher=SSL_get_current_cipher(ssl);
+
+	if (cipher)
+	{
+		const char *c;
+
+		c=SSL_CIPHER_get_version(cipher);
+		if (c)
+		{
+			(*dump_func)("Version: ", -1, dump_arg);
+			(*dump_func)(c, -1, dump_arg);
+			(*dump_func)("\n", 1, dump_arg);
+		}
+
+		{
+			char buf[10];
+
+			(*dump_func)("Bits: ", -1, dump_arg);
+
+			snprintf(buf, sizeof(buf), "%d",
+				 SSL_CIPHER_get_bits(cipher, NULL));
+			buf[sizeof(buf)-1]=0;
+
+			(*dump_func)(buf, -1, dump_arg);
+			(*dump_func)("\n", 1, dump_arg);
+		}
+
+		c=SSL_CIPHER_get_name(cipher);
+
+		if (c)
+		{
+			(*dump_func)("Cipher: ", -1, dump_arg);
+			(*dump_func)(c, -1, dump_arg);
+			(*dump_func)("\n", 1, dump_arg);
+		}
+	}
+}
+
+char *tls_get_encryption_desc(ssl_handle ssl)
+{
+	char protocolbuf[256];
+	SSL_CIPHER *cipher;
+	const char *c, *d;
+
+	cipher=SSL_get_current_cipher(ssl);
+
+	c=cipher ? SSL_CIPHER_get_version(cipher):NULL;
+	d=cipher ? SSL_CIPHER_get_name(cipher):NULL;
+
+	snprintf(protocolbuf, sizeof(protocolbuf),
+		 "%s,%dbits,%s",
+		 c ? c:"unknown",
+		 cipher ? SSL_CIPHER_get_bits(cipher, NULL):0,
+		 d ? d:"unknown");
+	protocolbuf[sizeof(protocolbuf)-1]=0;
+	return strdup(protocolbuf);
+}
+
+
+/* ------------------- */
+
+int tls_validate_pem_cert(const char *buf, size_t buf_size)
+{
+	int rc;
+	BIO *certbio;
+	int err;
+	EVP_PKEY *pk;
+	X509 *x;
+
+	ERR_clear_error();
+
+	rc=0;
+	certbio=BIO_new_mem_buf((void *)buf, buf_size);
+
+	if (!certbio)
+		return (0);
+
+	x=PEM_read_bio_X509(certbio, NULL, NULL, NULL);
+
+	if (x)
+	{
+		X509_free(x);
+
+		while ((x=PEM_read_bio_X509(certbio, NULL, NULL, NULL)) != NULL)
+			X509_free(x);
+
+		err = ERR_peek_last_error();
+                if (ERR_GET_LIB(err) == ERR_LIB_PEM &&
+		    ERR_GET_REASON(err) == PEM_R_NO_START_LINE)
+		{
+			rc=1;
+		}
+	}
+
+	ERR_clear_error();
+	BIO_free(certbio);
+
+	certbio=BIO_new_mem_buf((void *)buf, buf_size);
+
+	if (!certbio)
+		return (0);
+
+	if (!(pk=PEM_read_bio_PrivateKey(certbio, NULL, NULL, NULL)))
+	{
+		BIO_free(certbio);
+		ERR_clear_error();
+		return 0;
+	}
+
+	EVP_PKEY_free(pk);
+	return rc;
+}
+
+static size_t conv_name_to_rfc2553(const char *p, char *q)
+{
+#define PUTC(c) if (q) *q++=(c); ++n
+
+	size_t n=0;
+	const char *sep="";
+
+	while (*p)
+	{
+		if (*p == '/')
+		{
+			++p;
+			continue;
+		}
+
+		while (*sep)
+		{
+			PUTC(*sep);
+			++sep;
+		}
+		sep=",";
+
+		while (*p && *p != '/')
+		{
+			if (*p == '\\' && p[1])
+				++p;
+			if (*p == '\\' || *p == ',')
+			{
+				PUTC('\\');
+			}
+			PUTC(*p);
+			++p;
+		}
+	}
+	PUTC(0);
+#undef PUTC
+
+	return n;
+}
+
+char *tls_cert_name(const char *buf, size_t buf_size)
+{
+	int rc;
+	BIO *certbio;
+	char *p, *q;
+	X509 *x;
+	size_t cnt;
+
+	rc=0;
+	certbio=BIO_new_mem_buf((void *)buf, buf_size);
+
+	if (!certbio)
+	{
+		ERR_clear_error();
+		return (0);
+	}
+
+	x=PEM_read_bio_X509(certbio, NULL, NULL, NULL);
+	p=0;
+	q=0;
+
+	if (x)
+	{
+		p=X509_NAME_oneline(x->cert_info->subject, NULL, 0);
+		X509_free(x);
+	}
+	ERR_clear_error();
+	BIO_free(certbio);
+
+	if (p)
+	{
+		cnt=conv_name_to_rfc2553(p, NULL);
+
+		q=malloc(cnt);
+
+		if (q)
+			conv_name_to_rfc2553(p, q);
+		free(p);
+	}
+
+	return q;
+}
